-----------------------------------------------------------------------------
--
--  SPI_CMD.VHDL
--
--  Serializes and Deserializes a SPI-Data-Stream to Parallel Signals
--
--	Author 	: Rene Bllhoff
--  Created : around May 2005
--
--	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
-----------------------------------------------------------------------------
--
--  This source file may be used and distributed without
--  restriction provided that this copyright statement is not
--  removed from the file and that any derivative work contains
--  the original copyright notice and the associated disclaimer.
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--										
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public
--  License along with this source; if not, download it	from
--  http://www.gnu.org/licenses/gpl.html	
--
-----------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity spi_slave is
    generic (
      db_width : integer := 8
    );
    
    port (
	    clk50m  			: in  std_logic;
	    res     			: in  std_logic;        
    
	    spi_cs  			: in  std_logic;
	    spi_clk 			: in  std_logic;
	    spi_di  			: in  std_logic;
	    spi_do  			: out std_logic;  
		                      
			cs						: out std_logic;
	    dat2host      : in  std_logic_vector (db_width - 1 downto 0);
	    dat2dev       : out std_logic_vector (db_width - 1 downto 0);
			dat2dev_rcv   : out std_logic
	  );
end spi_slave;


         
architecture behave of spi_slave is
  
	-- state signal for spi state machine  
  signal spi_state  : std_logic_vector (2 downto 0);

	-- clocking in of spi_clk and spi_cs
  signal spi_clk_z1	: std_logic;
  signal spi_clk_z2	: std_logic;
  signal spi_clk_z3	: std_logic;
 	signal spi_cs_z1	: std_logic;
 	signal spi_cs_z2	: std_logic;
 	signal spi_cs_z3	: std_logic;

	-- receiving complete (and edge detection for trigger generation)	
	signal spi_rcvd		 : std_logic;
	signal spi_rcvd_z1 : std_logic;
	
	-- spi-clock-edge (0 = rising; 1 = falling)
	signal spi_ledge : std_logic;
    
  signal data_send  : std_logic_vector  (db_width downto 0); 
	signal data_recv  : std_logic_vector (db_width downto 0); 

	signal spi_busy		: std_logic;
	signal spi_host_ack : std_logic;

begin
   
  process (clk50m, res)
  begin
    -- RESET

    if res = '1' then
      
			-- reset clocking in signals
			spi_clk_z1  <= '1';
			spi_clk_z2  <= '1';
			spi_clk_z3  <= '1';
			spi_cs_z1   <= '1';
			spi_cs_z2   <= '1';
			spi_cs_z3   <= '1';

			spi_rcvd		<= '0';
			spi_rcvd_z1	<= '0';

			spi_do      <= '1';
			
			spi_busy		<= '0';
			spi_ledge   <= '0';
			
	    data_send 	<= (others => '0');
			data_recv 	<=  ( 0 => '1', others => '0');
											
			dat2dev     <= (others => '0');
			spi_state		<= (others => '0');

 	    -- extern reset       
        
      dat2dev_rcv <= '0';

      spi_do 			<= '1';  

		-- CLOCK

    elsif clk50m'event and clk50m = '1' then

      -- SPI HANDLING
			case spi_state is
				when "000"  => 	if spi_cs_z3 = '0' and spi_cs_z2 = '0' and spi_cs_z1 = '0' then
													spi_state		<= "001";
													data_recv		<= ( 0 => '1', others => '0');
												else
													spi_state 	<= "000";
													spi_do      <= '1';
													spi_busy    <= '0';
												end if;
			  when "001" 	=> 	if spi_clk_z3 = '0' and spi_clk_z2 = '0' and spi_clk_z1 = '1' then
													spi_state 	<= "010";	
													data_recv		<= data_recv (db_width - 1 downto 0) & spi_di;
													spi_rcvd		<= '0';	
													spi_busy		<= '1';
													spi_ledge  	<= '1';
												elsif spi_clk_z3 = '1' and spi_clk_z2 = '1' and spi_clk_z1 = '0' then
													spi_state 	<= "010";
													spi_do    	<= data_send (db_width - 1);
													spi_rcvd		<= '0';
													spi_busy		<= '1';
													spi_ledge  	<= '0';
												elsif spi_cs_z3 = '1' and spi_cs_z2 = '1' and spi_cs_z1 = '1' then
												  spi_state 	<= "000";
												end if;		 
											 
				when "010"  => 	if data_recv (db_width) = '1' and spi_ledge = '1' then
									   			spi_state 	<= "011";
												else
											   	spi_state 	<= "001";	 	
												end if;
												if spi_ledge = '0' then		 
	  							        data_send 	<= data_send (db_width - 1 downto 0) & '0';	
	  							      end if;  
												if spi_cs_z1 = '1' then
												   spi_state 	<= "000";
												end if;		 
				when "011"  => 	dat2dev     	<= data_recv (db_width - 1 downto 0); 
												spi_rcvd 			<= '1';
												spi_state 	 	<= "000";
												spi_busy			<= '0';												 
												if spi_cs_z1 = '1' then
													spi_state 	<= "000";
												end if;		 
				when others => spi_state   		<= "000";
			end case;

			-- clock detection
 		  spi_clk_z1 <= spi_clk;
 		  spi_clk_z2 <= spi_clk_z1;
 		  spi_clk_z3 <= spi_clk_z2;
			spi_cs_z1  <= spi_cs;
			spi_cs_z2  <= spi_cs_z1;
			spi_cs_z3  <= spi_cs_z2;

		  -- INTERFACE

 		  -- send request ?
			if spi_busy = '0' and spi_host_ack = '0' then
			  data_send (db_width - 1 downto 0)	<= dat2host;
				spi_host_ack 	<= '1';  	-- bernahme signalisieren
			else 
       	spi_host_ack 	<= '0';		-- aktiv oder keine daten		  
			end if;

			spi_rcvd_z1 <= spi_rcvd;
      dat2dev_rcv <= spi_rcvd and not spi_rcvd_z1;

  	end if;

  end process;  

	cs <= spi_cs_z3;

end architecture behave;