-----------------------------------------------------------------------------
--
--  DEVICE_6xBRAM_IOMEM_IF.VHDL
--
--  Implements an Interface for BlockRAM accesses via Data-Dispatcher
--  It offers 6-BlockRAM Interfaces (8-Bit DataPort) and a Memory Mapped IO
--  Interface ith 8 Bit Addr-Bus and 8 Bit Data-bus.
--  It Holds a statemachine to access (Read and Write) BlockRAM as well as 
--  the Memory mapped IO-Interface
--
--	Author 	: Rene Bllhoff
--  Created : around Apr. 2006
--
--	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
-----------------------------------------------------------------------------
--
--  This source file may be used and distributed without
--  restriction provided that this copyright statement is not
--  removed from the file and that any derivative work contains
--  the original copyright notice and the associated disclaimer.
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--										
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public
--  License along with this source; if not, download it	from
--  http://www.gnu.org/licenses/gpl.html	
--
-----------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity device_6xbram_iomem is
    Port ( clk 				: in 	std_logic;
           res 				: in 	std_logic;
					 
					 -- Dispatcher Interface
           fsm_trg 		: in 	std_logic;
           fsm_rdy 		: out std_logic;
           fsm_sel 		: in 	std_logic;
           fsm_din 		: in 	std_logic_vector ( 7 downto 0);
           fsm_dout 	: out std_logic_vector ( 7 downto 0);
					 fsm_inf		: out std_logic_vector (15 downto 0);
					 
					 -- BlockRAM Interface (common Signals Address, Data Write, Write Enable)
           bram_adr 		: out std_logic_vector(10 downto 0);
           bram_wdata		: out std_logic_vector( 7 downto 0);
           bram_wen 		: out std_logic;

					 -- BlockRAM 1 Interface (Read-Data and BlockRAM Enable)
           bram_rdata1	: in 	std_logic_vector( 7 downto 0);
           bram_en1			: out std_logic;
					 
					 -- BlockRAM 2 Interface (Read-Data and BlockRAM Enable)
           bram_rdata2	: in 	std_logic_vector( 7 downto 0);
           bram_en2			: out std_logic;
					 
					 -- BlockRAM 3 Interface (Read-Data and BlockRAM Enable)
           bram_rdata3	: in 	std_logic_vector( 7 downto 0);
           bram_en3			: out std_logic;
					 
					 -- BlockRAM 4 Interface (Read-Data and BlockRAM Enable)
           bram_rdata4	: in 	std_logic_vector( 7 downto 0);
           bram_en4			: out std_logic;

					 -- BlockRAM 5 Interface (Read-Data and BlockRAM Enable)
           bram_rdata5	: in 	std_logic_vector( 7 downto 0);
           bram_en5			: out std_logic;
					 
					 -- BlockRAM 6 Interface (Read-Data and BlockRAM Enable)
           bram_rdata6	: in 	std_logic_vector( 7 downto 0);
           bram_en6			: out std_logic;

					 -- Memory Mapped IO
					 iomem_adr		: out std_logic_vector (7 downto 0);
					 iomem_wdata	: out std_logic_vector (7 downto 0);
					 iomem_rdata	: in  std_logic_vector (7 downto 0);
					 iomem_wen		: out std_logic;
					 iomem_en			: out std_logic
					 );
end device_6xbram_iomem;

architecture Behavioral of device_6xbram_iomem is

	---------------------------------------------
	--
	-- Statemachine Types
	--
	---------------------------------------------	

	-- Device States
	type bram_states is  (bram_kind,				-- State : Access is Write Mem, Read Mem, Write IO, Read IO
												bram_mem_adr_hi,	-- State : Set Hi-Address Byte for Memory Access
												bram_mem_adr_lo,	-- State : Set Lo-Address Byte for Memory Access
												bram_mem_read,		-- State : Read Byte from BlockRAM
												bram_mem_write,		-- State : Write Byte to BlockRAM
												bram_iomem_adr,		-- State : Set Address Byte for IO Access
												bram_iomem_read,	-- State : Read Byte from IO
												bram_iomem_write	-- State : Write Byte to IO
										 	);

	-- Memory and IO Access States
	type bram_mstates is (bram_midle,				-- State : Idle, waiting all the time ...
												bram_mwrite,			-- State : Memory Write 
												bram_mwrite_01,		-- State : Memory Write 
												bram_mwrite_02,		-- State : Memory Write 
												bram_mwrite_03,	 	-- State : Memory Write 
												bram_mwrite_04,		-- State : Memory Write 
												bram_mread,			 	-- State : Memory Read 
												bram_mread_01,		-- State : Memory Read 	
												bram_mread_02,		-- State : Memory Read 
												bram_mread_03,		-- State : Memory Read 
												bram_iowrite, 		-- State : IO Write 
												bram_iowrite_01, 	-- State : IO Write 
												bram_iowrite_02, 	-- State : IO Write 
												bram_ioread, 			-- State : IO Read 
												bram_ioread_01, 	-- State : IO Read 
												bram_ioread_02	 	-- State : IO Read 
											);


	---------------------------------------------
	--
	-- 	Signals
	--
	---------------------------------------------	

	-- Statemachine Signals
	signal bram_state	 	: bram_states;
	signal bram_mstate 	: bram_mstates;


	-- Address Signals (for BlockRAM and IO)
	signal s_bram_adr   : std_logic_vector (15 downto 0);
	signal s_iomem_adr  : std_logic_vector ( 7 downto 0);

	-- Write Enable Signals (for BlockRAM and IO)
	signal s_bram_wen  	: std_logic;
	signal s_iomem_wen 	: std_logic;

	-- Global Muxer Enable
	signal bram_mena   	: std_logic;


begin
	

	---------------------------------------------
	--
	-- Output Signals
	--
	---------------------------------------------	

 
	bram_wen <= s_bram_wen;
	bram_adr <= s_bram_adr (10 downto 0);

	iomem_wen	<= s_iomem_wen;
	iomem_adr <= s_iomem_adr;


	---------------------------------------------
	--
	-- DEVICE_6&xBRAM_IOMEM Process
	--
	---------------------------------------------	

  process (clk, res)
	begin
		if res = '1' then

			bram_state	<= bram_kind;
			bram_mstate	<= bram_midle;
			s_bram_adr  <= (others => '0');   
			bram_wdata	<= (others => '0');   
			s_bram_wen  <= '0';  
			bram_mena		<= '0';									 
			fsm_rdy  		<= '1';  

		  iomem_wdata	<= (others => '0');
		  s_iomem_wen	<= '0';

		elsif clk'event and clk = '1' then



			---------------------------------------------
			--
			--  Main Statemachine
			--
			---------------------------------------------	
																																																	
			-- Generic Behaviour : only if fsm_sel = '1' enable triggering statemachine by fsm_trg = '1',
			-- 										 otherwise (when fsm_sel = '0') hold statemachine in reset
			if fsm_sel = '1' then
			  fsm_rdy <= '0';
				if fsm_trg = '1' then
				  -- Statemachine Memory Access Sequence
					case bram_state is
						when bram_kind => 
						  -- State bram_kind 				: Access Byte -> 0 = Memory Write 1 = Memory Read 2 = IO Write 3 = IO Read
							case fsm_din (3 downto 0) is
								-- Memory Write
								when X"0" 	=> s_bram_wen <= '1';
															 bram_state <= bram_mem_adr_hi;
															 fsm_rdy 		<= '0';
								-- Memory Read
								when X"1" 	=> s_bram_wen <= '0';
															 bram_state <= bram_mem_adr_hi;
															 fsm_rdy 		<= '0';
								-- IO Write
								when X"2" 	=> s_iomem_wen	<= '1';
															 bram_state <= bram_iomem_adr;
															 fsm_rdy 		<= '0';
								-- IO Read
								when X"3" 	=> s_iomem_wen 	<= '0';
															 bram_state <= bram_iomem_adr;
															 fsm_rdy 		<= '0';
								when others => NULL;
							end case;

					  -- State bram_mem_adr_hi 	:	Copy Byte to High-Byte of Memory Address Register
						when bram_mem_adr_hi =>
						  s_bram_adr (15 downto 8) <= fsm_din;
						 	bram_state 	<= bram_mem_adr_lo;											 
						 	fsm_rdy 		<= '0';

					  -- State bram_mem_adr_lo 	:	Copy Byte to Low-Byte of Memory Address Register
						when bram_mem_adr_lo =>
						  s_bram_adr ( 7 downto 0) <= fsm_din;
						 	fsm_rdy 		<= '0';
							-- if write -> go to write state
							if s_bram_wen = '1' then
								bram_state 	<= bram_mem_write;
							else
								-- if read -> go to read state and trigger memory read
								bram_state  <= bram_mem_read;
								bram_mstate <= bram_mread;
							end if;

						-- State bram_mem_write 	:	Copy Byte to Memory Data Out Register and Trigger Memory Write
						when bram_mem_write =>
							bram_wdata 	<= fsm_din;
							bram_mstate <= bram_mwrite;
							-- Remain in this state to allow autoincrement function
							fsm_rdy 		<= '0';	 -- REMBEMER : fsm_rdy = '0' means -> this state will never be left, exceopt bringing fsm_sel down

						-- State bram_mem_read 		:	Trigger Memory Read (byte is autmatically copied to DEVICE_6xBRAM_IOMEM Output Register)
						when bram_mem_read =>
							bram_mstate <= bram_mread;
							-- Remain in this state to allow autoincrement function
							fsm_rdy 		<= '0'; -- REMBEMER : fsm_rdy = '0' means -> this state will never be left, exceopt bringing fsm_sel down


						-- State bram_iomem_adr 	:	Copy Byte to Memory Mapped IO Address Register
						when bram_iomem_adr =>
						  s_iomem_adr <= fsm_din;
						 	fsm_rdy 		<= '0';
							-- if write -> go to write io state
							if s_iomem_wen = '1' then
								bram_state 	<= bram_iomem_write;
							-- if read -> go to io read state and trigger io read
							else
								bram_state  <= bram_iomem_read;
								bram_mstate <= bram_ioread;
							end if;

						-- State bram_iomem_write :	Copy Byte to IO Data Output and Trigger IO Write
						when bram_iomem_write =>
							iomem_wdata <= fsm_din;
							bram_mstate <= bram_iowrite;
							fsm_rdy 		<= '1';

						-- State bram_iomem_read 	:	Trigger IO Read (byte is autmatically copied to DEVICE_6xBRAM_IOMEM Output Register)
						when bram_iomem_read =>
							fsm_rdy 		<= '1';

						when others => NULL;
					end case;				  
				end if;
			else
				-- Generic behaviour : if fsm_sel = '0' hold statemachine in reset
				bram_state 	<= bram_kind;
				fsm_rdy 		<= '1';
				bram_mstate	<= bram_midle;
			end if;


			---------------------------------------------
			--
			-- Statemachine for Accessing BlockRAM and IO-Memory
			--
			---------------------------------------------	
	
			-- bram_mstate is set by Main Statemachine (within this process) 
			case bram_mstate is
				-- State bram_midle 		: Idle, Do Nothing
			  when bram_midle => 
					bram_mena 	<= '0';
					iomem_en 		<= '0';

				-- State bram_mwrite 		: Entry Point Write BlockRAM Memory (enable Data In Muxer & Enable Muxer)
				when bram_mwrite =>
					bram_mena 	 <= '1';
				  bram_mstate <= bram_mwrite_01;
				-- State bram_mwrite_01 : Wait cycle
				when bram_mwrite_01 =>
					bram_mena 	 <= '1';
				  bram_mstate <= bram_mwrite_02;
				-- State bram_mwrite_02 : Wait cycle 
				when bram_mwrite_02 =>
					bram_mena 	 <= '0';
					bram_mstate <= bram_mwrite_03;
				-- State bram_mwrite_03 : Increment Memory Address Register and go to idle 
				when bram_mwrite_03 =>
					bram_mena 	 <= '0';
					s_bram_adr		 <= s_bram_adr + '1';
					bram_mstate <= bram_midle;

				-- State bram_mread 		: Entry Point Read BlockRAM Memory (enable Enable & Data In Muxer)
				when bram_mread =>
				  bram_mena 	 <= '1';
				  bram_mstate <= bram_mread_01;
				-- State bram_mread_01 	: Wait cycle 
				when bram_mread_01 =>
				  bram_mena 	 <= '1';
				  bram_mstate <= bram_mread_02;
				-- State bram_mread_02 	: Copy Data In from Device to Data Out Register
				when bram_mread_02 =>
					bram_mena 	<= '1';
				  bram_mstate <= bram_mread_03;
					-- dependend on Memory Address Copy Data In to Output Register
					case s_bram_adr (15 downto 12) is
						when X"0" 	=> fsm_dout <= bram_rdata1;
						when X"2" 	=> fsm_dout <= bram_rdata2;
						when X"4" 	=> fsm_dout <= bram_rdata3;
						when X"6" 	=> fsm_dout <= bram_rdata4;
						when X"8" 	=> fsm_dout <= bram_rdata5;
						when X"a" 	=> fsm_dout <= bram_rdata6;
						when others => NULL;
					end case;
				-- State bram_mread_03 	: Increment Memory Address Register and go to idle  
				when bram_mread_03 =>
				  bram_mena 	 <= '0';
					s_bram_adr		 <= s_bram_adr + '1';
				  bram_mstate <= bram_midle;


				-- State bram_iowrite 	: Entry Point Write IO Memory and Set IO Enable
				when bram_iowrite =>
				  iomem_en 		<= '0';
				  bram_mstate <= bram_iowrite_01;
				-- State bram_iowrite_01 : Wait cycle 
				when bram_iowrite_01 =>
				  iomem_en 		<= '1';
				  bram_mstate <= bram_iowrite_02;
				-- State bram_iowrite_02 : Take Back IO Enable and go to idle
				when bram_iowrite_02 =>
					iomem_en 		<= '0';
					s_iomem_adr <= s_iomem_adr + '1';
					bram_mstate <= bram_midle;


				-- State bram_ioread 		: Entry Point Read IO Memory
				when bram_ioread =>
				  iomem_en 	 	<= '1';
				  bram_mstate <= bram_ioread_01;
				-- State bram_ioread_01 : Wait Cycle
				when bram_ioread_01 =>
				  iomem_en 	 	<= '0';
				  bram_mstate <= bram_ioread_02;
				-- State bram_ioread_02 : Copy Data In from Device to Data Out Register
				when bram_ioread_02 =>
					fsm_dout 		<= iomem_rdata;
					iomem_en 	 	<= '0';
					s_iomem_adr <= s_iomem_adr + '1';
					bram_mstate <= bram_midle;

				when others => NULL;

			end case;
	  end if;


		---------------------------------------------
		--
		-- 	Enable Muxer and Data In Muxer
		--
		---------------------------------------------	
		
		-- If Memory Enable is set
		if bram_mena = '1' then
		  -- dependend on Memory Address set specific Enable BlockRAM	Signals
			case s_bram_adr (15 downto 12) is
				when X"0"		=> bram_en1 <= '1'; bram_en2 <= '0'; bram_en3 <= '0'; bram_en4 <= '0'; bram_en5 <= '0'; bram_en6 <= '0';
				when X"2"		=> bram_en1 <= '0'; bram_en2 <= '1'; bram_en3 <= '0'; bram_en4 <= '0'; bram_en5 <= '0'; bram_en6 <= '0';
				when X"4"		=> bram_en1 <= '0'; bram_en2 <= '0'; bram_en3 <= '1'; bram_en4 <= '0'; bram_en5 <= '0'; bram_en6 <= '0';
				when X"6"		=> bram_en1 <= '0'; bram_en2 <= '0'; bram_en3 <= '0'; bram_en4 <= '1'; bram_en5 <= '0'; bram_en6 <= '0';
				when X"8"		=> bram_en1 <= '0'; bram_en2 <= '0'; bram_en3 <= '0'; bram_en4 <= '0'; bram_en5 <= '1'; bram_en6 <= '0';
				when X"a"		=> bram_en1 <= '0'; bram_en2 <= '0'; bram_en3 <= '0'; bram_en4 <= '0'; bram_en5 <= '0'; bram_en6 <= '1';
				when others => bram_en1 <= '0'; bram_en2 <= '0'; bram_en3 <= '0'; bram_en4 <= '0'; bram_en5 <= '0'; bram_en6 <= '0';
			end case;
		else
		  -- if no Memory Enable take back all Enable BlockRAM Signals
			bram_en1 <= '0'; bram_en2 <= '0'; bram_en3 <= '0'; bram_en4 <= '0'; bram_en5 <= '0'; bram_en6 <= '0';
		end if;

	end process;

end Behavioral;
